/****************************************************************************
 * randroot - a program to take a list of images and randomly display them
 *            over time.
 *
 ****************************************************************************
 *
 * Version 1.0  Scott Powers (spowers@ncsa.uiuc.edu)
 *	Loads configuration file, works
 *
 * Version 1.1  Brandon Long (blong@uiuc.edu)
 *	- Added command line specification of where the configuration file is
 *	- Added ability to kill another if already running, and log pid 
 *        to aid it
 *      - slight code rearrangement
 *
 */

/**************************** System include files ************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/utsname.h>
#include <string.h>
#include <sys/types.h>
#include <signal.h>

/**************************** Global #defines ****************************/ 
#define VERSION "1.1"
#define RCFILE ".randrootrc"
#define PIDFILE ".randroot.pid"
#define MODE_RANDOM 0
#define MODE_ASCENDING_SEQUENTIAL 1
#define MODE_DESCENDING_SEQUENTIAL 2
#define MODE_RANDOM_SEQUENTIAL 3

#define FALSE 0
#define TRUE  1

/**************************** Global Variables ****************************/ 
int numImages=0;
char **images=NULL;
char *displayCommand=NULL;
int timeDelay=(-1);
int displayMode=(-1);
char rcFileName[BUFSIZ];
char pidFileName[BUFSIZ];
int quickQuit=0;


/**************************** Function headers ****************************/
void addImage(char *src);
void readConfigFile();
char *makeword(char *line, char stop);
int makeFilename();
void run_loop();
void kill_file(char *pid_fname);
void log_pid(char *pid_fname);


int usage(char *argv0) 
{
    fprintf(stderr,"\nRandroot %s: Random Root Image Chooser\n",VERSION);
    fprintf(stderr,"\tScott Powers (spowers@ncsa.uiuc.edu)");
    fprintf(stderr,"\tBrandon Long (blong@uiuc.edu)\n\n");
    fprintf(stderr,"Usage: %s [-vkq] [-r rcfilename] [-p pidfilename]\n",
            argv0);
    fprintf(stderr,"\t-v\t\t: Echo version and usage information\n");
    fprintf(stderr,"\t-k\t\t: Kill currently running randroot\n");
    fprintf(stderr,"\t-q\t\t: Quit after displaying one image\n");
    fprintf(stderr,"\t-r rcfilename\t: Specify alternate Configuration File\n");
    fprintf(stderr,"\t-p pidfilename\t: Specify alternate PID File\n\n");
    exit(1);
}

int main(int argc, char **argv) 
{
    char *image_cmd=NULL;
    int index,up,down,direction;
    int c,kill_current;
    extern char *optarg;
    extern int optind;
    
    rcFileName[0] = '\0';
    pidFileName[0] = '\0';
    
    while((c = getopt(argc,argv,"qkr:p:vh")) != -1) {
	switch(c) {
	case 'q':
	    quickQuit = TRUE;
	    break;
	case 'r':
	    strcpy(rcFileName,optarg);
	    break;
        case 'k':
            kill_current = TRUE;
            break;
        case 'p':
            strcpy(pidFileName,optarg);
            break;
        case 'h':
        case 'v':
	case '?':
	    usage(argv[0]);
	    break;
	}
    }
    
    if (!makeFileName()) {
        fprintf(stderr,"Could not find HOME directory to read %s or %s.\n",
                RCFILE,PIDFILE);
        fprintf(stderr,"Use the -r and/or -p arguments to specify where ");
        fprintf(stderr,"to find the file\n");
        usage(argv[0]);
        exit(1);
    }

    if (kill_current == TRUE) {
      kill_file(pidFileName);
    }
    
    log_pid(pidFileName);

    readConfigFile();
    
    if (!check_config()) exit(1);
 
    run_loop();
 
    exit(0);
}

/* Checks various globals to see if they have been set by parsing the 
 * configuration file.  Returns 1 if everything is ok, or 0 if something
 * needs to be fixed.
 */
int check_config()
{
    int ret = 1;

    if (!displayCommand) {
        fprintf(stderr,"You do not have a display command specified in %s.\n",
                rcFileName);
        ret = 0;
    }
 
    if (timeDelay<=0) {
        fprintf(stderr,"You do not have a time delay specified in %s.\n",
                rcFileName);
        ret = 0;
    }
 
    if (displayMode<0) {
        fprintf(stderr,"You do not have a display mode specified in %s.\n",
                rcFileName);
        ret = 0;
    }
 
    if (numImages<1) {
        fprintf(stderr,"You do not have any images specified in %s.\n",
                rcFileName);
        ret = 0;
    }
 
    if (numImages==1) {
        fprintf(stderr,"WARNING: You have only specified 1 image! Kinda \n");
        fprintf(stderr,"  defeats the point of this program...doesn't it?\n");
        ret = 0;
    }
    return ret;
}

void run_loop() 
{
    int index, direction,down,up;
    char *image_cmd;

    image_cmd=(char *)calloc(BUFSIZ+strlen(displayCommand)+512,sizeof(char));
 
    srand48(time(0));
 
    index=((int)(drand48()*100000)%numImages);
    down=index-1;
    if (down<0) {
        down=numImages-1;
    }
    up=index;

    for (;;) {
        if (displayMode==MODE_RANDOM) {
            index=((int)(drand48()*100000)%numImages);
        }
        else if (displayMode==MODE_ASCENDING_SEQUENTIAL) {
            index++;
            if (index>=numImages) {
                index=0;
            }
        }
        else if (displayMode==MODE_DESCENDING_SEQUENTIAL) {
            index--;
            if (index<0) {
                index=numImages-1;
            }
        }
        else if (displayMode==MODE_RANDOM_SEQUENTIAL) {
            direction=((int)(drand48()*100000)%100);
            if (direction<50) { /*down*/
                down--;
                if (down<0) {
                    down=numImages-1;
                }
                if (down==up) {
                    down--;
                    if (down<0) {
                        down=numImages-1;
                    }
                }
                index=down;
            }
            else { /*up*/
                up++;
                if (up>=numImages) {
                    up=0;
                }
                if (up==down) {
                    up++;
                    if (up>=numImages) {
                        up=0;
                    }
                }
                index=up;
            }
        }
        sprintf(image_cmd,"%s %s",displayCommand,images[index]);
        system(image_cmd);
        if (quickQuit) {
            break;
        }
        sleep(timeDelay);
    }
}

void kill_file(char *pid_fname)
{
   FILE *pid_file;
   pid_t pid;

   if(!(pid_file = fopen(pid_fname,"r"))) {
        fprintf(stderr,"randroot: could not open pid log %s\n",pid_fname);
        perror("fopen");
        exit(1);
    }
    fscanf(pid_file,"%d\n",&pid);
    fclose(pid_file); 
 
    if (kill(pid,SIGTERM)) {
      fprintf(stderr,"randroot: cound not kill process %d\n",pid);
      perror("fopen");
      exit(1);
    }
}

void log_pid(char *pid_fname)
{
    FILE *pid_file;
    
    if(!(pid_file = fopen(pid_fname,"w"))) {
        fprintf(stderr,"randroot: could not log pid to file %s\n",pid_fname);
        perror("fopen");
        exit(1);
    }
    fprintf(pid_file,"%d\n",getpid());
    fclose(pid_file);
}

void readConfigFile() 
{
    FILE *fp;
    char *token,str[BUFSIZ],*ptr;
    
    if (!(fp=fopen(rcFileName,"r"))) {
	perror(rcFileName);
	
	printf("\n");
	printf("Would you like me to write out a default %s so you can\n",
               rcFileName);
	printf("  see what it looks like and how it works (suggested!)? [DEFAULT=Y] ");
     
	{
	    char dummy[80];
	    
	    gets(dummy);
	    
	    if (!*dummy || (*dummy && tolower(*dummy)=='y')) {
		if (!(fp=fopen(rcFileName,"w"))) {
		    perror(rcFileName);
		    exit(1);
		}
		fprintf(fp,"#\n");
		fprintf(fp,"# Resource command file for randroot!\n");
		fprintf(fp,"#\n");
		fprintf(fp,"# Written by: Scott Powers [10/11/1995]\n");
		fprintf(fp,"# spowers@ncsa.uiuc.edu\n");
		fprintf(fp,"#\n");
		fprintf(fp,"\n");
		fprintf(fp,"# All of the following examples are commented out. You can use a\n");
		fprintf(fp,"#   # sign at the start of a line for a comment or an ! point.\n");
		fprintf(fp,"#   Regardless, you _must_ specify a Display, Time Delay, and at\n");
		fprintf(fp,"#   least 1 image.\n");
		fprintf(fp,"\n");
		fprintf(fp,"# DISPLAY COMMAND\n");
		fprintf(fp,"#   The Display Command is the command used to put the image onto\n");
		fprintf(fp,"#   your root window. You may use anything you like... xv works fine.\n");
		fprintf(fp,"#\n");
		fprintf(fp,"# The following is what I use:\n");
		fprintf(fp,"#display: /opt/bin/xv -root -quit -rmode 5\n");
		fprintf(fp,"\n");
		fprintf(fp,"# TIME DELAY COMMAND\n");
		fprintf(fp,"#  The Time Delay Command is the number of seconds to wait before\n");
		fprintf(fp,"#  displaying the next random image.\n");
		fprintf(fp,"#\n");
		fprintf(fp,"# The following is what I use:\n");
		fprintf(fp,"#time: 300\n");
		fprintf(fp,"\n");
		fprintf(fp,"# MODE COMMAND\n");
		fprintf(fp,"#   The Mode Command is the way in which images are displayed. You can\n");
		fprintf(fp,"#   choose from: random, ascending-sequential, descending-sequential, or\n");
		fprintf(fp,"#   random-sequential. Random uses drand48 and is seeded with time(0),\n");
		fprintf(fp,"#   thus the images are chosen at pseudo-randomness. Ascending-sequential\n");
		fprintf(fp,"#   is from image 0 to image n in sequential order. Descending-sequential is\n");
		fprintf(fp,"#   from image n to image 0 in sequential order. Random-sequential uses\n");
		fprintf(fp,"#   drand48 to pick a starting image number, then uses drand48 to go up or\n");
		fprintf(fp,"#   down each time progressing farther and farther (when the \"up\" and \"down\"\n");
		fprintf(fp,"#   meet, they skip over each other).\n");
		fprintf(fp,"#\n");
		fprintf(fp,"# The following is what I use:\n");
		fprintf(fp,"#mode: random-sequential\n");
		fprintf(fp,"\n");
		fprintf(fp,"# IMAGE COMMAND\n");
		fprintf(fp,"#   The Image Command is used to specify the image to display to the\n");
		fprintf(fp,"#   root window. You are expected to have more than one since that is\n");
		fprintf(fp,"#   the purpose of this program.\n");
		fprintf(fp,"#   NOTE1: You must specify the _full_ path to the image. This means no\n");
		fprintf(fp,"#     tilda support is in place.\n");
		fprintf(fp,"#   NOTE2: Randroot automatically appends the image filename, so do not\n");
		fprintf(fp,"#     specify one!\n");
		fprintf(fp,"#\n");
		fprintf(fp,"# The following are 5 of the many images I use:\n");
		fprintf(fp,"#image: /sam/htdocs/Tolkien/TolkPics/lotr/tolk-balrog2.gif\n");
		fprintf(fp,"#image: /sam/htdocs/Tolkien/TolkPics/lotr/tolk-rivendell.gif\n");
		fprintf(fp,"#image: /sam/htdocs/Tolkien/TolkPics/lotr/tolk-cliff.gif\n");
		fprintf(fp,"#image: /sam/htdocs/Tolkien/TolkPics/lotr/tolk-greenhill.gif\n");
		fprintf(fp,"#image: /sam/htdocs/Tolkien/TolkPics/lotr/tolk-ford2.gif\n");
		fprintf(fp,"\n");
		fprintf(fp,"# The End.\n");
		
		fclose(fp);
		
		printf("\n\nNow you need to edit %s to configure randroot for your system!\n",rcFileName);
		exit(1);
	    }
	    else {
		printf("Okay...well you _must_ have a %s to use randroot!\n",
                       rcFileName);
		exit(1);
	    }
	}
	return;
    }
    
    while (!feof(fp)) {
	fgets(str,BUFSIZ-1,fp);
	ptr=str;
	
	if (feof(fp)) {
	    continue;
	}
	
	for (; ptr && isspace(*ptr); ptr++);
	
	if (!ptr || !*ptr) { /*blank line*/
	    continue;
	}
	
	if (*ptr=='#' || *ptr=='!') { /*comment*/
	    continue;
	}
	
	token=makeword(ptr,':');
	for (; ptr && isspace(*ptr); ptr++);
	for (; token && isspace(*token); token++);
	if (!ptr || !*ptr) {
	    fprintf(stderr,"No value associated with the token [%s]\n",token);
	    continue;
	}
	*(ptr+strlen(ptr)-1)='\0';
	
	if (token && *token) {
	    if (!strncasecmp(token,"image",5)) {
		addImage(ptr);
		continue;
	    }
	    if (!strncasecmp(token,"display",6)) {
		displayCommand=strdup(ptr);
		continue;
	    }
	    if (!strncasecmp(token,"time",4)) {
		timeDelay=atoi(ptr);
		continue;
	    }
	    if (!strncasecmp(token,"mode",4)) {
		if (!strcasecmp(ptr,"random")) {
		    displayMode=MODE_RANDOM;
		}
		else if (!strcasecmp(ptr,"ascending-sequential")) {
		    displayMode=MODE_ASCENDING_SEQUENTIAL;
		}
		else if (!strcasecmp(ptr,"descending-sequential")) {
		    displayMode=MODE_DESCENDING_SEQUENTIAL;
		}
		else if (!strcasecmp(ptr,"random-sequential")) {
		    displayMode=MODE_RANDOM_SEQUENTIAL;
		}
		else {
		    fprintf(stderr,"Error in value [%s] associated with the token [%s].\n  Please choose from: random, ascending-sequential, descending-sequential,\n  or random-sequential.",ptr,token);
		    fprintf(stderr,"Defaulting to: random\n");
		    displayMode=MODE_RANDOM;
		}
		continue;
	    }
	}
	
	fprintf(stderr,"Weird error in %s. Attempting to continue.\n",rcFileName);
	continue;
    }
    
    return;
}


void addImage(char *src) 
{
    FILE *fp;
    
    if (!src && !*src) {
	return;
    }
    
    if (!(fp=fopen(src,"r"))) {
	perror(src);
	return;
    }
    fclose(fp);
    
    if (!numImages) {
	images=(char **)calloc(1,sizeof(char *));
    }
    else {
	images=(char **)realloc(images,((numImages+1)*sizeof(char *)));
    }
    
    if (!images) {
	perror("calloc/realloc");
	exit(1);
    }
    
    images[numImages]=strdup(src);
    if (!images[numImages]) {
	perror("strdup");
	exit(1);
    }
    
    numImages++;
    
    return;
}


char *makeword(char *line, char stop) 
{
    int x = 0,y;
    char *word = (char *) malloc(sizeof(char) * (strlen(line) + 1));
    
    for(x=0;((line[x]) && (line[x] != stop));x++)
        word[x] = line[x];
    
    word[x] = '\0';
    if(line[x]) ++x;
    y=0;
    
    while(line[y++] = line[x++]);
    return word;
}


int makeFileName() 
{
    char *hptr, home[256];
    struct passwd *pwdent;

    /* If we already got an rcFileName and pidFileName, just return 
     */
    if (rcFileName[0] && pidFileName[0]) return 1;
    
    /*
     * Try the HOME environment variable, then the password file, and
     *   finally give up.
     */
    if (!(hptr=getenv("HOME"))) {
	if (!(pwdent=getpwuid(getuid()))) {
	    return(0);
	}
	else {
	    strcpy(home,pwdent->pw_dir);
	}
    }
    else {
	strcpy(home,hptr);
    }
    
    if (!rcFileName[0]) sprintf(rcFileName,"%s/%s",home,RCFILE);
    if (!pidFileName[0]) sprintf(pidFileName,"%s/%s",home,PIDFILE);
    
    return(1);
}
